/*
 * Copyright (c) 2002 Jason W. Bacon All rights reserved.
 *
 * Redistribution and use in source and binary forms, with or without
 * modification, are permitted provided that the following conditions are
 * met: 1. Redistributions of source code must retain the above copyright
 * notice, this list of conditions and the following disclaimer. 2.
 * Redistributions in binary form must reproduce the above copyright notice,
 * this list of conditions and the following disclaimer in the documentation
 * and/or other materials provided with the distribution.
 *
 * THIS SOFTWARE IS PROVIDED BY THE AUTHOR AND CONTRIBUTORS ``AS IS'' AND ANY
 * EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE IMPLIED
 * WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE ARE
 * DISCLAIMED.  IN NO EVENT SHALL THE AUTHOR OR CONTRIBUTORS BE LIABLE FOR
 * ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL
 * DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR
 * SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER
 * CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT
 * LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY
 * OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF
 * SUCH DAMAGE.
 *
 */

#include <stdio.h>
#include <string.h>
#include <ctype.h>
#include <fcntl.h>
#include <stdlib.h>
#include <unistd.h>
#include <fnmatch.h>
#include <errno.h>
#include <sys/stat.h>
#include <sys/ioctl.h>
#include <sys/cdio.h>
#include <sysexits.h>
#include <sys/param.h>
#include <sys/ucred.h>
#include <sys/mount.h>
#include "eject.h"
#include "protos.h"

int     main(int argc, char *argv[])

{
    flag_t  flags;
    char    *device;

    /* Check command line options */
    switch(flags = get_flags(&argc,argv))
    {
	case    USAGE_ERROR:
	    usage(argv);
	default:
	    break;
    }
    
    if ( HELP(flags) || INVALID(flags) )
	usage(argv);

    /* If all is well, device is specified by the last command line arg */
    /* or the default if none was given */
    if ( USE_DEFAULT(flags) )
	device = DEFAULT_DEVICE;
    else
	if ( (device = normalize_devname(argv[argc-1],argv,flags)) == NULL )
	    return EX_NOINPUT;

    if ( ! device_allowed(device,argv[0]) )
    {
	fprintf(stderr,"%s: Ejecting %s is not allowed.\n",argv[0],device);
	return EX_NOPERM;
    }
	
    /* Unmount device from all mount points */
    unmount_all(device,argv,flags);
    
    /* Eject */
    return control_dev(device,argv,flags);
}


flag_t  get_flags(int *argc,char *argv[])

{
    int     a;
    flag_t  flags = FLAG_INIT;
    char    *p;
    
    for (a=1; (a<*argc) && (*argv[a] == '-'); ++a)
    {
	p = argv[a] + 1;
	if ( *p == '-' )    /* GNU style -- flag */
	{
	    flags |= gnu_flag(p+1);
	}
	else                /* Bundled flags */
	{
	    while ( *p != '\0' )
	    {
		flags |= bundled_flag(*p++);
	    }
	}
    }
    
    /* Should be one argument left - the device or mount point name */
    if ( a == *argc-1 )
	return flags;
    else if ( a == *argc )
    {
	if ( VERBOSE(flags) )
	    printf("Using default device: %s.\n",DEFAULT_DEVICE);
	flags |= USE_DEFAULT_FLAG;
	return flags;
    }
    else
	return USAGE_ERROR;
}


flag_t  gnu_flag(char *p)

{
    if ( strcmp(p,"verbose") == 0 )
	return VERBOSE_FLAG;
    if ( strcmp(p,"noop") == 0 )
	return NOOP_FLAG;
    if ( strcmp(p,"help") == 0 )
	return HELP_FLAG;
    if ( strcmp(p,"trayclose") == 0 )
	return CLOSE_TRAY_FLAG;
    if ( strcmp(p,"locktray") == 0 )
	return LOCK_TRAY_FLAG;
    else
    {
	fprintf(stderr,"Invalid GNU flag: --%s\n",p);
	return INVALID_FLAG;
    }
}


flag_t  bundled_flag(int ch)

{
    switch(ch)
    {
	case    'h':
	case    '?':
	    return HELP_FLAG;
	case    'v':
	    return VERBOSE_FLAG;
	case    'n':
	    return NOOP_FLAG;
	case    't':
	    return CLOSE_TRAY_FLAG;
	case    'l':
	    return LOCK_TRAY_FLAG;
	default:
	    fprintf(stderr,"Invalid bundled flag: -%c\n",ch);
	    return INVALID_FLAG;
    }
}


/**********************************************************************
    Convert user-specified device name or mount point to a device
    name suitable for opening with open().
 **********************************************************************/

char    *normalize_devname(char *name, char *argv[],flag_t flags)

{
    FILE   *fstab;
    char    buff[STR_MAX + 1],
	    dev[STR_MAX + 1],
	    mnt[STR_MAX + 1],
	    *bp;
    static char device[STR_MAX+1] = "";

    /* If user specified a full-path, it must be either /dev/something
       or a mount point */
    if (*name == '/')
    {
	/* If not /dev, scan fstab for a matching mount point */
	if (memcmp(name, "/dev", 4) != 0)
	{
	    if ((fstab = fopen("/etc/fstab", "r")) == NULL)
	    {
		fputs("Cannot open fstab...\n", stderr);
	    }
	    while ((bp = fgets(buff, STR_MAX, fstab)) != NULL)
	    {
		sscanf(buff, "%s %s", dev, mnt);
		if ((memcmp(dev, "/dev", 4) == 0) && (strcmp(mnt, name) == 0))
		{
		    if ( VERBOSE(flags) )
			printf("From /etc/fstab: %s %s\n", dev, mnt);
		    strlcpy(device, dev, STR_MAX);
		    break;  /* Quit loop when match found */
		}
	    }
	    fclose(fstab);
	    if ( bp == NULL )
	    {
		fprintf(stderr,"%s: Error: %s not found in fstab.\n",
			argv[0],name);
		return NULL;
	    }
	}
	
	/* Massage user-specified name into proper device for control_dev() */
	else
	{
	    strlcpy(device, name, STR_MAX);
	}
    }
    else    /* No leading / - assume a device basename, like "acd0" */
    {
	/* CD device name must end in a letter */
	if (isdigit(name[strlen(name) - 1]))
	    /* User left off letter suffix, e.g. acd0 */
	    snprintf(device, STR_MAX, "/dev/%sc", name);
	else
	    /* User included letter suffix, e.g. acd0 */
	    snprintf(device, STR_MAX, "/dev/%s", name);
    }
    if ( VERBOSE(flags) )
	printf("%s: using device %s\n", argv[0], device);
    return device;
}


/**********************************************************************
    Eject or otherwise control the device
 **********************************************************************/

int     control_dev(char *device, char *argv[], int flags)

{
    int     fd, status;
    extern int  errno;
    struct ioc_capability;
    
    if (!NOOP(flags))
    {
	if ( (fd = open(device, O_RDWR)) == -1 )
	{
	    fprintf(stderr,"%s: Open failed: %s.\n",
		    argv[0],strerror(errno));
	    return OPEN_FAILED;
	}
	/*ioctl(fd,CDIOCCAPABILITY,&cd_cap);*/
	/* Required for some drives with disk loaded. */
	status = ioctl(fd, CDIOCALLOW); 
	if ( 1 )
	{
	    if ( NON_EJECT(flags) )
	    {
		if ( CLOSE_TRAY(flags) )
		{
		    if ( VERBOSE(flags) )
			printf("%s: Closing tray: %s\n",argv[0],device);
		    status = ioctl(fd, CDIOCCLOSE);
		}
		if ( LOCK_TRAY(flags) )
		{
		    if ( VERBOSE(flags) )
			printf("%s: Locking tray: %s\n",argv[0],device);
		    status = ioctl(fd, CDIOCPREVENT);
		}
	    }
	    else
	    {
		if ( VERBOSE(flags) )
		    printf("%s: Ejecting %s\n",argv[0],device);
		status = ioctl(fd, CDIOCEJECT);
	    }
	}
	else
	{
	    fprintf(stderr,"%s: CDIOCALLOW failed: %s.\n",
		    argv[0],strerror(errno));
	    return CDIO_NOT_ALLOWED;
	}
	close(fd);
    }
    else
    {
	printf("%s: Ejecting %s\n", argv[0], device);
	status = 0;
    }
    return status;
}


void    usage(char *argv[])

{
    static char *gnu_opts[] = VALID_GNU_OPTS,
		**p;
    
    fprintf(stderr, "usage: %s [-%s|GNU style flags] [device|mount point]\n", argv[0], VALID_OPTS);
    puts("    GNU style flags:");
    for (p=gnu_opts; *p != NULL; ++p)
	printf("    %s\n",*p);
    exit(1);
}


void    unmount_all(char *device,char *argv[],flag_t flags)

{
    struct statfs *mounts;
    int     m,c;
    
    /* Get a list of all currently mounted devices */
    m = getmntinfo(&mounts,MNT_WAIT);
    
    /* Scan list for device and unmount all matching mount points */
    for (c=0; c<m; ++c)
    {
	if ( strcmp(device,mounts[c].f_mntfromname) == 0 ) 
	{
	    if ( VERBOSE(flags) )
		printf("%s is currently mounted.  Unmounting...\n",
		    mounts[c].f_mntonname);
	    if ( unmount(mounts[c].f_mntonname,0) == -1 )
	    {
		fprintf(stderr,"%s: Cannot unmount %s: %s\n",
		    argv[0],device,strerror(errno));
		exit(EX_OSERR);
	    }
	}
    }
}


/**********************************************************************
    Check /usr/local/etc/eject.allow for device name.  Return
    TRUE if the device is allowed (listed in the file), FALSE
    otherwise.
 **********************************************************************/

int     device_allowed(char *device,char *prog_name)

{
    FILE    *fp;
    static char    *fname = "/usr/local/etc/eject.allow";
    char    line[STR_MAX+1];
    
    fp = fopen(fname,"r");
    if ( fp == NULL )
    {
	fprintf(stderr,"Cannot open %s.  Make sure the file exists\n",fname);
	fputs("and that the eject command has suid permissions.\n",stderr);
	return FALSE;
    }
    
    while ( fgets(line,STR_MAX,fp) != NULL )
    {
	/* Strip comments and blank lines */
	if ( (*line != '#') && !strblank(line) ) 
	{
	    /* Clip trailing newline before fnmatch() */
	    line[strlen(line)-1] = '\0';    
	    if ( fnmatch(line,device,0) == 0 )
	    {
		fclose(fp);
		return TRUE;
	    }
	}
    }
    fclose(fp);
    return FALSE;
}


int     strblank(str)
char    *str;

{
    while ( *str != '\0' )
    {
	if ( !isspace(*str) )   /* Not blank? */
	    return 0;
	++str;
    }
    return 1;
}

